use errors::Result;
use mail_store::{create_mail_store, MailStore, Name};
use mail_store_state::{MailFolder, MailStoreState};
use message::{Message, Uid};
use message_threads::MessageThreader;
use std::ops::DerefMut;
use std::path::PathBuf;
use std::sync::mpsc::{channel, Receiver, SendError, Sender};
use std::sync::Arc;
use std::thread;
use sync_tree::TreeChange;

pub trait FoldersReceiver: Send + 'static {
    fn set_folders(&self, Vec<Name>);
}

pub trait FolderReceiver: Send + 'static {
    fn add_changes(&mut self, changes: Vec<TreeChange<Arc<Message>>>);
}

pub trait EmailReceiver: Send + 'static {
    fn set_email(&self, Option<Arc<Message>>);
}

pub enum Job<R: FoldersReceiver, F: FolderReceiver, E: EmailReceiver> {
    SetConfig(R, PathBuf),
    SetFolder(F, String),
    SetEmail(E, String, Uid),
    Stop,
}

struct JobThread<R: FoldersReceiver, F: FolderReceiver, E: EmailReceiver> {
    job_rx: Receiver<Job<R, F, E>>,
    mail_store: Option<Box<MailStore>>,
    current_folder: Option<(F, Option<Arc<MailFolder>>, String)>,
    last_folder: Option<String>,
    state: MailStoreState,
}

impl<R: FoldersReceiver, F: FolderReceiver, E: EmailReceiver> JobThread<R, F, E> {
    fn mail_store(&mut self) -> Result<&mut MailStore> {
        if let Some(mail_store) = &mut self.mail_store {
            Ok(mail_store.deref_mut())
        } else {
            bail!("No mail store was initiated.")
        }
    }
    fn is_folder_complete(&self) -> bool {
        self.current_folder
            .as_ref()
            .map(|(_, _, f)| {
                if let Some(mail_store) = &self.mail_store {
                    mail_store.is_folder_complete(f, &self.state)
                } else {
                    false
                }
            }).unwrap_or(true)
    }
    fn start(job_receiver: Receiver<Job<R, F, E>>) -> thread::JoinHandle<()> {
        thread::spawn(move || {
            let mut job_thread = JobThread {
                job_rx: job_receiver,
                mail_store: None,
                current_folder: None,
                last_folder: None,
                state: MailStoreState::default(),
            };
            let mut r = job_thread.job_rx.recv().ok();
            loop {
                r = job_thread.thread_loop(r);
                if r.is_none() {
                    return;
                }
            }
        })
    }
    fn thread_loop(&mut self, r: Option<Job<R, F, E>>) -> Option<Job<R, F, E>> {
        let r = match r {
            None | Some(Job::Stop) => {
                println!("stopping mail model thread");
                return None;
            }
            Some(Job::SetConfig(mut folders, config_path)) => {
                self.state = MailStoreState::default();
                folders.set_folders(Vec::new());
                match create_mail_store(&config_path) {
                    Ok(mail_store) => {
                        self.mail_store = Some(mail_store);
                        let _ = self.list_folders(&folders);
                        Ok(())
                    }
                    Err(err) => {
                        self.mail_store = None;
                        Err(err)
                    }
                }
            }
            Some(Job::SetFolder(mut folder, folder_name)) => {
                let f = self.state.get_folder(&folder_name);
                let mut changes = Vec::new();
                if let Some(f) = &f {
                    changes.push(TreeChange::Reset(f.threads.tree().clone()));
                } else {
                    changes.push(TreeChange::Reset(Arc::default()));
                }
                folder.add_changes(changes);
                self.current_folder = Some((folder, f, folder_name));
                Ok(())
            }
            Some(Job::SetEmail(email, folder, uid)) => self.set_email(&email, &folder, uid),
        };
        if let Err(e) = r {
            println!("Error for command: {}", e);
            for e in e.iter().skip(1) {
                println!("caused by: {}", e);
            }
        }
        while !self.is_folder_complete() {
            if let Ok(job) = self.job_rx.try_recv() {
                return Some(job);
            }
            let (mut ui, folder, folder_name) = self.current_folder.take().unwrap();
            match self.list_folder(&mut ui, folder, &folder_name) {
                Err(e) => println!("Error for command: {}", e),
                Ok(folder) => {
                    self.state
                        .set_folder(folder_name.clone(), Arc::clone(&folder));
                    self.current_folder = Some((ui, Some(folder), folder_name));
                }
            }
        }
        self.job_rx.recv().ok()
    }
    fn list_folders(&mut self, receiver: &R) -> Result<()> {
        let folders = self.mail_store()?.list_folders()?;
        receiver.set_folders(folders);
        Ok(())
    }
    fn list_folder(
        &mut self,
        ui: &mut F,
        folder: Option<Arc<MailFolder>>,
        folder_path: &str,
    ) -> Result<Arc<MailFolder>> {
        let threads = if let Some(folder) = folder {
            folder.threads.clone()
        } else {
            Arc::default()
        };
        let new_folder = Some(folder_path) == self.last_folder.as_ref().map(|a| a.as_str());

        let mut threader = MessageThreader::new(&threads, threads.tree().clone(), !new_folder);
        self.mail_store()?
            .fetch_headers(folder_path, &mut threader)?;
        let (changes, threads) = threader.done();
        ui.add_changes(changes);
        Ok(Arc::new(MailFolder::new(Arc::new(threads))))
    }
    fn set_email(&mut self, email: &E, folder: &str, uid: Uid) -> Result<()> {
        let message = self.mail_store()?.fetch(folder, uid)?;
        email.set_email(Some(Arc::new(message)));
        Ok(())
    }
}

pub struct JobProcessor<R: FoldersReceiver, F: FolderReceiver, E: EmailReceiver> {
    job_tx: Sender<Job<R, F, E>>,
    handle: Option<thread::JoinHandle<()>>,
}

impl<R: FoldersReceiver, F: FolderReceiver, E: EmailReceiver> JobProcessor<R, F, E> {
    pub fn new() -> JobProcessor<R, F, E> {
        let (job_tx, job_rx) = channel();
        let handle = JobThread::start(job_rx);
        JobProcessor {
            job_tx,
            handle: Some(handle),
        }
    }
    pub fn send(&self, job: Job<R, F, E>) -> ::std::result::Result<(), SendError<Job<R, F, E>>> {
        self.job_tx.send(job)
    }
}

impl<R: FoldersReceiver, F: FolderReceiver, E: EmailReceiver> Drop for JobProcessor<R, F, E> {
    fn drop(&mut self) {
        let _ = self.job_tx.send(Job::Stop);
        self.job_tx = channel().0; // close the channel by dropping it
        self.handle.take().map(|handle| handle.join());
    }
}
